@@ -50,22 +50,30 @@ edit_dynamic_rep::find_dynamic (path p) {
 ******************************************************************************/
 
 bool
-edit_dynamic_rep::is_multi_paragraph_macro (tree t) {
+edit_dynamic_rep::is_multi_paragraph_macro (tree t, bool pure) {
   int n= arity (t);
   if (is_document (t) || is_func (t, PARA) || is_func (t, SURROUND))
     return true;
   if (is_func (t, MACRO) || is_func (t, WITH) ||
       is_func (t, LOCUS) ||
-      is_func (t, CANVAS) || is_func (t, ORNAMENT))
-    return is_multi_paragraph_macro (t [n-1]);
-  if (is_extension (t) && (!is_compound (t, "footnote"))) {
+      is_func (t, CANVAS) || is_func (t, ORNAMENT) || is_func (t, ART_BOX))
+    return is_multi_paragraph_macro (t [n-1], pure);
+  if (is_extension (t) &&
+      !is_compound (t, "footnote") &&
+      !is_compound (t, "footnote-anchor") &&
+      !is_compound (t, "note-footnote") &&
+      !is_compound (t, "note-footnote*") &&
+      !is_compound (t, "folded-comment")) {
     int i;
     for (i=1; i<n; i++)
-      if (is_multi_paragraph_macro (t[i]))
-	return true;
-    tree f= get_env_value (t[0]->label);
-    return is_multi_paragraph_macro (f);
+      if (is_multi_paragraph_macro (t[i], pure))
+        return true;
+    tree f= get_env_value (as_string (L(t)));
+    return is_multi_paragraph_macro (f, pure);
   }
+  if (!pure)
+    if (is_func (t, ARG))
+      return true;
   return false;
 }
 
@@ -77,7 +85,7 @@ contains_table_format (tree t, tree var) {
     int i, n= N(t);
     for (i=0; i<n; i++)
       if (contains_table_format (t[i], var))
-	return true;
+        return true;
     return is_func (t, TFORMAT) && (t[N(t)-1] == tree (ARG, var));
   }
 }
@@ -89,7 +97,7 @@ edit_dynamic_rep::make_compound (tree_label l, int n= -1) {
   if (n == -1) {
     for (n=0; true; n++) {
       if (drd->correct_arity (l, n) &&
-	  ((n>0) || (drd->get_arity_mode (l) == ARITY_NORMAL))) break;
+          ((n>0) || (drd->get_arity_mode (l) == ARITY_NORMAL))) break;
       if (n == 100) return;
     }
   }
@@ -104,24 +112,39 @@ edit_dynamic_rep::make_compound (tree_label l, int n= -1) {
   if (n == 0) insert_tree (t, 1);
   else if (is_with_like (t) && as_bool (call ("with-like-check-insert", t)));
   else {
-    tree f= get_env_value (as_string (l));
-    bool block_macro= (N(f) == 2) && is_multi_paragraph_macro (f);
+    string s= as_string (l);
+    tree f= get_env_value (s);
+    bool block_macro= (N(f) == 2) && is_multi_paragraph_macro (f, true);
+    bool large_macro= (N(f) == 2) && is_multi_paragraph_macro (f, false);
     bool table_macro= (N(f) == 2) && contains_table_format (f[1], f[0]);
+    // FIXME: why do we take the precaution N(f) == 2 ?
+    if (s == "explain") block_macro= true;
 
     tree sel= "";
     if (selection_active_small () ||
-	(block_macro && selection_active_normal ()))
+        (large_macro && selection_active_normal ()))
       sel= selection_get_cut ();
-    if ((block_macro && (!table_macro)) ||
-	(l == make_tree_label ("footnote")))
-      {
-	t[0]= tree (DOCUMENT, "");
-	p   = path (0, 0, 0);
-      }
-    if ((!drd->all_accessible (l)) && get_init_string (MODE) != "src") {
-      t= tree (INACTIVE, t);
-      p= path (0, p);
+    else if (is_with_like (t) && selection_active_normal ()) {
+      sel= selection_get_cut ();
+      t[n-1]= sel;
+      insert_tree (t, p);
+      return;
     }
+    if ((block_macro && (!table_macro)) ||
+        (l == make_tree_label ("footnote")) ||
+        (l == make_tree_label ("footnote-anchor")) ||
+        (l == make_tree_label ("note-footnote")) ||
+        (l == make_tree_label ("note-footnote*")) ||
+        (l == make_tree_label ("folded-comment")))
+      {
+        t[0]= tree (DOCUMENT, "");
+        p   = path (0, 0, 0);
+      }
+    if (!drd->all_accessible (l))
+      if (get_init_string (MODE) != "src" && !inside ("show-preamble")) {
+        t= tree (INACTIVE, t);
+        p= path (0, p);
+      }
     insert_tree (t, p);
     if (table_macro) make_table (1, 1);
     if (sel != "") insert_tree (sel, end (sel));
@@ -148,12 +171,14 @@ edit_dynamic_rep::activate () {
     tree u (make_tree_label (st[0]->label));
     u << A (st (1, N(st)));
     assign (p, u);
+    call ("notify-activated", object (subtree (et, p)));
     go_to (end (et, p));
     correct (path_up (p));
   }
   else {
     bool acc= (p < path_up (tp) && drd->is_accessible_child (st, tp[N(p)]));
     remove_node (p * 0);
+    call ("notify-activated", object (subtree (et, p)));
     if (!acc) go_to (end (et, p));
     correct (path_up (p));
   }
@@ -172,10 +197,11 @@ edit_dynamic_rep::go_to_argument (path p, bool start_flag) {
   else if (i >= n) go_to_end (path_up (p, inactive? 2: 1));
   else {
     if ((!drd->is_accessible_child (t, i)) &&
-	(!inactive) && (!in_source ()))
+        (!inactive) && (!in_source ()))
       {
-	insert_node (path_up (p) * 0, INACTIVE);
-	p= path_up (p) * path (0, i);
+        insert_node (path_up (p) * 0, INACTIVE);
+        call ("notify-disactivated", object (subtree (et, path_up (p) * 0)));
+        p= path_up (p) * path (0, i);
       }
     if (start_flag) go_to_start (p);
     else go_to_end (p);
@@ -220,26 +246,26 @@ edit_dynamic_rep::remove_empty_argument (path p, bool forward) {
 
   for (d=1; d<=n-i; d++)
     if ((src_flag && (d==1)) ||
-	((!src_flag) &&
-	 drd->correct_arity (L(t), n-d) &&
-	 drd->insert_point (L(t), i, n-d)))
+        ((!src_flag) &&
+         drd->correct_arity (L(t), n-d) &&
+         drd->insert_point (L(t), i, n-d)))
       {
-	bool flag= true;
-	for (j=0; j<d; j++)
-	  flag= flag && is_empty (t[i+j]);
-	if (flag) {
-	  bool old_locked= env_locked; env_locked= true;
-	  remove (p, d);
-	  if ((d == n) && is_mod_active_once (subtree (et, path_up (p, 2)))) {
-	    remove_node (path_up (p, 2) * 0);
-	    go_to_border (path_up (p, 2), forward);
-	  }
-	  else if (forward) go_to_argument (path_up (p) * i, true);
-	  else go_to_argument (path_up (p) * (i-1), false);
-	  env_locked= old_locked;
-	  return;
-	}
-	else break;
+        bool flag= true;
+        for (j=0; j<d; j++)
+          flag= flag && is_empty (t[i+j]);
+        if (flag) {
+          bool old_locked= env_locked; env_locked= true;
+          remove (p, d);
+          if ((d == n) && is_mod_active_once (subtree (et, path_up (p, 2)))) {
+            remove_node (path_up (p, 2) * 0);
+            go_to_border (path_up (p, 2), forward);
+          }
+          else if (forward) go_to_argument (path_up (p) * i, true);
+          else go_to_argument (path_up (p) * (i-1), false);
+          env_locked= old_locked;
+          return;
+        }
+        else break;
       }
 
   bool flag= true;
@@ -249,22 +275,24 @@ edit_dynamic_rep::remove_empty_argument (path p, bool forward) {
     assign (path_up (p), "");
     tree st= subtree (et, path_up (p, 2));
     if ((is_mod_active_once (st) || is_compound (st, "doc-inactive")) &&
-	(st[0] == "")) {
+        (st[0] == "")) {
       assign (path_up (p, 2), "");
       correct (path_up (p, 3));
     }
     // FIXME: temporary hack for doc-data and doc-author-data
-    else if (is_compound (st, "doc-data") ||
-	     is_compound (st, "doc-author-data")) {
+    else if (is_compound (st, "doc-data")      ||
+             is_compound (st, "abstract-data") ||
+             is_compound (st, "doc-author")    ||
+             is_compound (st, "author-data")) {
       if (N(st)==1) {
-	assign (path_up (p, 2), "");
-	correct (path_up (p, 3));
+        assign (path_up (p, 2), "");
+        correct (path_up (p, 3));
       }
       else {
-	int i= last_item (path_up (p)) + (forward? 0: -1);
-	remove (path_up (p), 1);
-	if (i<0) go_to_start (path_up (p, 2));
-	else go_to_border (path_up (p, 2) * i, forward);
+        int i= last_item (path_up (p)) + (forward? 0: -1);
+        remove (path_up (p), 1);
+        if (i<0) go_to_start (path_up (p, 2));
+        else go_to_border (path_up (p, 2) * i, forward);
       }
     }
     else correct (path_up (p, 2));
@@ -327,7 +355,7 @@ edit_dynamic_rep::back_general (path p, bool forward) {
   }
   else if (n==0) back_monolithic (p);
   else if ((n==1) && is_func (st[0], DOCUMENT, 1) &&
-	   (is_func (st[0][0], TFORMAT) || is_func (st[0][0], TABLE)))
+           (is_func (st[0][0], TFORMAT) || is_func (st[0][0], TABLE)))
     back_table (p * path (0, 0), forward);
   else if ((n==1) && (is_func (st[0], TFORMAT) || is_func (st[0], TABLE)))
     back_table (p * 0, forward);
@@ -396,8 +424,8 @@ edit_dynamic_rep::insert_with (path p, string var, tree val) {
     int i, n= N(st)-1;
     for (i=0; i<n; i+=2)
       if (st[i] == var) {
-	assign (p * (i+1), copy (val));
-	return;
+        assign (p * (i+1), copy (val));
+        return;
       }
     insert (p * n, copy (tree (WITH, var, val)));    
   }
@@ -413,9 +441,9 @@ edit_dynamic_rep::remove_with (path p, string var) {
     int i, n= N(st)-1;
     for (i=0; i<n; i+=2)
       if (st[i] == var) {
-	remove (p * i, 2);
-	if (n == 2) remove_node (p * 0);
-	return;
+        remove (p * i, 2);
+        if (n == 2) remove_node (p * 0);
+        return;
       }
   }
   else if ((rp < p) && is_func (subtree (et, path_up (p)), WITH))
@@ -462,8 +490,8 @@ edit_dynamic_rep::insert_style_with (path p, string var, string val) {
     int i, n= N(st);
     for (i=n-1; i>=0; i-=2)
       if (st[i] == var) {
-	assign (path_up (p) * (i+1), copy (val));
-	return;
+        assign (path_up (p) * (i+1), copy (val));
+        return;
       }
     insert (path_up (p) * (n-1), tree (STYLE_WITH, copy (var), copy (val)));
   }
@@ -500,7 +528,7 @@ edit_dynamic_rep::make_hybrid () {
   if (in_source ()) insert_tree (t, p);
   else insert_tree (tree (INACTIVE, t), path (0, p));
   set_message (concat (kbd ("return"), ": activate symbol or macro"),
-	       "hybrid");
+               "hybrid");
 }
 
 bool
@@ -521,7 +549,7 @@ edit_dynamic_rep::activate_latex () {
       return true;
     }
     set_message ("Error: not a command name",
-		 "activate latex command");
+                 "activate latex command");
   }
   return false;
 }
@@ -541,15 +569,16 @@ edit_dynamic_rep::activate_hybrid (bool with_args_hint) {
   // activate macro argument
   string name= st[0]->label;
   path mp= search_upwards (MACRO);
+  if (is_nil (mp)) mp= search_upwards ("edit-macro");
   if (!is_nil (mp)) {
     tree mt= subtree (et, mp);
     int i, n= N(mt)-1;
     for (i=0; i<n; i++)
       if (mt[i] == name) {
-	assign (p, tree (ARG, copy (name)));
-	go_to (end (et, p));
-	correct (path_up (p));
-	return;
+        assign (p, tree (ARG, copy (name)));
+        go_to (end (et, p));
+        correct (path_up (p));
+        return;
       }
   }
 
@@ -575,7 +604,7 @@ edit_dynamic_rep::activate_hybrid (bool with_args_hint) {
     if (N(st) == 2) insert_tree (st[1]);
   }
   else set_message ("Error: unknown command",
-		    "activate hybrid command");
+                    "activate hybrid command");
   env_locked= old_locked;
 }
 
@@ -597,7 +626,7 @@ edit_dynamic_rep::activate_symbol () {
     int i, n= N(s);
     for (i=0; i<n; i++)
       if ((s[i]=='<') || (s[i]=='>'))
-	{ s= ""; break; }
+        { s= ""; break; }
     assign (p, "<" * s * ">");
   }
   go_to (end (et, p));
