@@ -19,6 +19,7 @@
 #include <QMenuBar>
 #include <QLayoutItem>
 
+#include "config.h"
 #include "analyze.hpp"
 #include "scheme.hpp"
 
@@ -42,24 +43,29 @@ int menu_count = 0;  // zero if no menu is currently being displayed
 list<qt_tm_widget_rep*> waiting_widgets;
 
 static void
-replaceActions (QWidget* dest, QWidget* src) {
-    //NOTE: the parent hierarchy of the actions is not modified while installing
-    //      the menu in the GUI (see qt_menu.hpp for this memory management 
-    //      policy)
+replaceActions (QWidget* dest,  QList<QAction*>* src) {
+  //NOTE: the parent hierarchy of the actions is not modified while installing
+  //      the menu in the GUI (see qt_menu.hpp for this memory management
+  //      policy)
   if (src == NULL || dest == NULL)
     FAILED ("replaceActions expects valid objects");
   dest->setUpdatesEnabled (false);
   QList<QAction *> list = dest->actions();
-  while (!list.isEmpty()) dest->removeAction (list.takeFirst());
-  list = src->actions();
-  while (!list.isEmpty()) dest->addAction (list.takeFirst());
+  for (int i = 0; i < list.count(); i++) {
+    QAction* a = list[i];
+    dest->removeAction (a);
+  }
+  for (int i = 0; i < src->count(); i++) {
+    QAction* a = (*src)[i];
+    dest->addAction(a);
+  }
   dest->setUpdatesEnabled (true);
 }
 
 static void
-replaceButtons (QToolBar* dest, QWidget* src) {
-   if (src == NULL || dest == NULL)
-     FAILED ("replaceButtons expects valid objects");
+replaceButtons (QToolBar* dest, QList<QAction*>* src) {
+  if (src == NULL || dest == NULL)
+    FAILED ("replaceButtons expects valid objects");
   dest->setUpdatesEnabled (false);
   bool visible = dest->isVisible();
   if (visible) dest->hide(); //TRICK: to avoid flicker of the dest widget
@@ -69,7 +75,8 @@ replaceButtons (QToolBar* dest, QWidget* src) {
     QToolButton* button = qobject_cast<QToolButton*> (list[i]);
     if (button) {
       button->setPopupMode (QToolButton::InstantPopup);
-      button->setStyle (qtmstyle());
+      if (tm_style_sheet == "")
+        button->setStyle (qtmstyle());
     }
   }
   if (visible) dest->show(); //TRICK: see above
@@ -115,14 +122,31 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   // general setup for main window
   
   QMainWindow* mw= mainwindow ();
-  mw->setStyle (qtmstyle ());
-  mw->menuBar()->setStyle (qtmstyle ());
-  
+  if (tm_style_sheet == "") {
+    mw->setStyle (qtmstyle ());
+    mw->menuBar()->setStyle (qtmstyle ());
+  }
+
+#ifdef Q_OS_MAC
+  if (!use_native_menubar) {
+    mw->menuBar()->setNativeMenuBar(false);
+    if (tm_style_sheet != "") {
+      int min_h= (int) floor (28 * retina_scale);
+      mw->menuBar()->setMinimumHeight (min_h);
+    }
+  }
+#else
+  if (tm_style_sheet != "") {
+    int min_h= (int) floor (28 * retina_scale);
+    mw->menuBar()->setMinimumHeight (min_h);
+  }
+#endif
+
   // there is a bug in the early implementation of toolbars in Qt 4.6
   // which has been fixed in 4.6.2 (at least)
   // this is why we change dimension of icons
   
-#if (defined(Q_WS_MAC)&&(QT_VERSION>=QT_VERSION_CHECK(4,6,0))&&(QT_VERSION<QT_VERSION_CHECK(4,6,2)))
+#if (defined(Q_OS_MAC)&&(QT_VERSION>=QT_VERSION_CHECK(4,6,0))&&(QT_VERSION<QT_VERSION_CHECK(4,6,2)))
   mw->setIconSize (QSize (22, 30));  
 #else
   mw->setIconSize (QSize (17, 17));
@@ -136,9 +160,11 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   rightLabel= new QLabel (qt_translate ("Booting"), mw);
   leftLabel->setFrameStyle (QFrame::NoFrame);
   rightLabel->setFrameStyle (QFrame::NoFrame);
+  leftLabel->setIndent (8);
   bar->addWidget (leftLabel, 1);
   bar->addPermanentWidget (rightLabel);
-  bar->setStyle (qtmstyle ());
+  if (tm_style_sheet == "")
+    bar->setStyle (qtmstyle ());
   
   // NOTE (mg): the following setMinimumWidth command disable automatic 
   // enlarging of the status bar and consequently of the main window due to 
@@ -153,6 +179,25 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   //    trying to figure this out :)
   
   bar->setMinimumWidth (2);
+#ifdef Q_OS_LINUX
+  int min_h= (int) floor (28 * retina_scale);
+  bar->setMinimumHeight (min_h);
+#else
+#if (QT_VERSION >= 0x050000)
+  if (tm_style_sheet != "") {
+    int min_h= (int) floor (28 * retina_scale);
+    bar->setMinimumHeight (min_h);
+  }
+#else
+  double status_scale=
+    (((double) retina_icons) > retina_scale? 1.5: retina_scale);
+  if (status_scale > 1.0) {
+    int std_h= (os_mingw ()? 28: 20);
+    int min_h= (int) floor (std_h * status_scale);
+    bar->setMinimumHeight (min_h);
+  }
+#endif
+#endif
   mw->setStatusBar (bar);
  
   // toolbars
@@ -171,12 +216,14 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   dock_window_widget = tm_new<qt_window_widget_rep> (sideTools, dock_name,
                                                      command(), true);
   
-  mainToolBar->setStyle (qtmstyle ());
-  modeToolBar->setStyle (qtmstyle ());
-  focusToolBar->setStyle (qtmstyle ());
-  userToolBar->setStyle (qtmstyle ());
-  sideTools->setStyle (qtmstyle ());
-  bottomTools->setStyle (qtmstyle ());
+  if (tm_style_sheet == "") {
+    mainToolBar->setStyle (qtmstyle ());
+    modeToolBar->setStyle (qtmstyle ());
+    focusToolBar->setStyle (qtmstyle ());
+    userToolBar->setStyle (qtmstyle ());
+    sideTools->setStyle (qtmstyle ());
+    bottomTools->setStyle (qtmstyle ());
+  }
   
   {
     // set proper sizes for icons
@@ -192,10 +239,63 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
     sz = (pxm ? pxm->size() : QSize (16, 16));
     tweak_iconbar_size (sz);
     focusToolBar->setIconSize (sz);
-  }  
+  }
+
+  // Why we need fixed height:
+  // The height of the toolbar is actually determined by the font height.
+  // And the font height is not fixed. If the height of the toolbar is not
+  // fixed, the stretching of it will make the document area floating and
+  // triggers the re-rendering of the full document.
+  //
+  // NOTICE: setFixedHeight must be after setIconSize
+  // TODO: the size of the toolbar should be calculated dynamically
+#if (QT_VERSION >= 0x050000)
+#if defined (Q_OS_MAC) || defined (Q_OS_WIN)
+  int toolbarHeight= 30 * retina_icons;
+  mainToolBar->setFixedHeight (toolbarHeight + 8 * retina_icons);
+  modeToolBar->setFixedHeight (toolbarHeight + 4 * retina_icons);
+  focusToolBar->setFixedHeight (toolbarHeight);
+#else
+  int toolbarHeight= 30;
+  mainToolBar->setFixedHeight (toolbarHeight + 8);
+  modeToolBar->setFixedHeight (toolbarHeight + 4);
+  focusToolBar->setFixedHeight (toolbarHeight);
+#endif
+#else
+#ifdef Q_OS_MAC
+  if (retina_icons > 1) {
+    int toolbarHeight= 30;
+    if (!use_unified_toolbar)
+      mainToolBar->setFixedHeight (toolbarHeight + 8);
+    modeToolBar->setFixedHeight (toolbarHeight + 4);
+    focusToolBar->setFixedHeight (toolbarHeight);
+  }
+#else
+  int toolbarHeight= 30 * retina_icons;
+  mainToolBar->setFixedHeight (toolbarHeight + 8);
+  modeToolBar->setFixedHeight (toolbarHeight + 4);
+  focusToolBar->setFixedHeight (toolbarHeight);  
+#endif
+#endif
+  if (tm_style_sheet != "") {
+    double scale= retina_scale;
+#if ((QT_VERSION < 0x050000) && defined (Q_OS_MAC))
+    scale= max (scale, 0.6 * ((double) retina_icons));
+#endif
+    int h1= (int) floor (38 * scale + 0.5);
+    int h2= (int) floor (34 * scale + 0.5);
+    int h3= (int) floor (30 * scale + 0.5);
+#if ((QT_VERSION < 0x050000) && defined (Q_OS_MAC))
+    if (use_unified_toolbar && retina_icons == 2 && scale == 1.2) {
+      h1= 34; h2= 36; h3= 32; }
+#endif
+    mainToolBar->setFixedHeight (h1);
+    modeToolBar->setFixedHeight (h2);
+    focusToolBar->setFixedHeight (h3);
+  }
   
   QWidget *cw= new QWidget();
-  cw->setObjectName("central widget");  // this is important for styling toolbars.
+  cw->setObjectName("centralWidget");  // this is important for styling toolbars.
   
     // The main layout
   
@@ -209,51 +309,70 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   
   mw->setCentralWidget (cw);
 
+  mainToolBar->setObjectName ("mainToolBar");
+  modeToolBar->setObjectName ("modeToolBar");
+  focusToolBar->setObjectName ("focusToolBar");
+  userToolBar->setObjectName ("userToolBar");
+  bottomTools->setObjectName ("bottomTools");
+  sideTools->setObjectName ("sideTools");
+
 #ifdef UNIFIED_TOOLBAR
 
-  mw->setUnifiedTitleAndToolBarOnMac(true);
+  if (use_unified_toolbar) {
+    mw->setUnifiedTitleAndToolBarOnMac(true);
     
-  //WARNING: dumbToolBar is the toolbar installed on the top area of the
-  //main widget which is  then unified in the title bar. 
-  //to overcome some limitations of the unified toolbar implementation we
-  //install the real toolbars as widgets in this toolbar.
+    //WARNING: dumbToolBar is the toolbar installed on the top area of the
+    //main widget which is  then unified in the title bar. 
+    //to overcome some limitations of the unified toolbar implementation we
+    //install the real toolbars as widgets in this toolbar.
   
-  dumbToolBar = mw->addToolBar("dumb toolbar");
-  dumbToolBar->setMinimumHeight(30);
+    dumbToolBar = mw->addToolBar("dumb toolbar");
+    dumbToolBar->setMinimumHeight(30);
 
-  //these are the actions related to the various toolbars to be installed in
-  //the dumb toolbar.
+    //these are the actions related to the various toolbars to be installed in
+    //the dumb toolbar.
   
-  mainToolBarAction = dumbToolBar->addWidget(mainToolBar);
-  modeToolBarAction = NULL;
+    mainToolBarAction = dumbToolBar->addWidget(mainToolBar);
+    modeToolBarAction = NULL;
 
   
-  // A ruler
-  rulerWidget = new QWidget(cw);
-  rulerWidget->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
-  rulerWidget->setMinimumHeight(1);
-  rulerWidget->setBackgroundRole(QPalette::Mid);
-  // FIXME: how to use 112 (active) and 146 (passive)
-  rulerWidget->setVisible(false);
-  rulerWidget->setAutoFillBackground(true);
-  // rulerWidget = new QLabel("pippo", cw);
+    // A ruler
+    rulerWidget = new QWidget(cw);
+    rulerWidget->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
+    rulerWidget->setMinimumHeight(1);
+    rulerWidget->setBackgroundRole(QPalette::Mid);
+    // FIXME: how to use 112 (active) and 146 (passive)
+    rulerWidget->setVisible(false);
+    rulerWidget->setAutoFillBackground(true);
+    // rulerWidget = new QLabel("pippo", cw);
   
     // A second ruler (this one always visible) to separate from the canvas.
-  QWidget* r2 = new QWidget(mw);
-  r2->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
-  r2->setMinimumHeight(1);
-  r2->setBackgroundRole(QPalette::Mid);
-  r2->setVisible(true);
-  r2->setAutoFillBackground(true);
+    QWidget* r2 = new QWidget(mw);
+    r2->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Fixed);
+    r2->setMinimumHeight(1);
+    r2->setBackgroundRole(QPalette::Mid);
+    r2->setVisible(true);
+    r2->setAutoFillBackground(true);
 
-  bl->insertWidget(0, modeToolBar);
-  bl->insertWidget(1, rulerWidget);
-  bl->insertWidget(2, focusToolBar);
-  bl->insertWidget(3, userToolBar);
-  bl->insertWidget(5, r2);
+    bl->insertWidget(0, modeToolBar);
+    bl->insertWidget(1, rulerWidget);
+    bl->insertWidget(2, focusToolBar);
+    bl->insertWidget(3, userToolBar);
+    bl->insertWidget(5, r2);
 
     //mw->setContentsMargins (-2, -2, -2, -2);  // Why this?
-  bar->setContentsMargins (0, 1, 0, 1);
+    bar->setContentsMargins (0, 1, 0, 1);
+  }
+  else {
+    mw->addToolBar (mainToolBar);
+    mw->addToolBarBreak ();
+    mw->addToolBar (modeToolBar);
+    mw->addToolBarBreak ();
+    mw->addToolBar (focusToolBar);
+    mw->addToolBarBreak ();
+    mw->addToolBar (userToolBar);
+    mw->addToolBarBreak ();
+  }
 
 #else
   mw->addToolBar (mainToolBar);
@@ -270,6 +389,7 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   sideTools->setFeatures (QDockWidget::DockWidgetMovable |
                          QDockWidget::DockWidgetFloatable);
   sideTools->setFloating (false);
+  sideTools->setTitleBarWidget (new QWidget()); // Disables title bar
   mw->addDockWidget (Qt::RightDockWidgetArea, sideTools);
 
   bottomTools->setAllowedAreas (Qt::BottomDockWidgetArea);
@@ -299,9 +419,13 @@ qt_tm_widget_rep::qt_tm_widget_rep(int mask, command _quit)
   sideTools->setVisible (false);
   bottomTools->setVisible (false);
   mainwindow()->statusBar()->setVisible (true);
-#ifndef Q_WS_MAC
+#ifndef Q_OS_MAC
   mainwindow()->menuBar()->setVisible (false);
-#endif  
+#endif
+  QPalette pal;
+  QColor bgcol= to_qcolor (tm_background);
+  pal.setColor (QPalette::Mid, bgcol);
+  mainwindow()->setPalette(pal);
 }
 
 qt_tm_widget_rep::~qt_tm_widget_rep () {
@@ -315,6 +439,19 @@ qt_tm_widget_rep::~qt_tm_widget_rep () {
 
 void
 qt_tm_widget_rep::tweak_iconbar_size (QSize& sz) {
+#ifdef Q_OS_LINUX
+  if (sz.height () >= 24) {
+    sz.setWidth (sz.width () + 2);
+    sz.setHeight (sz.height () + 8);
+  }
+  else if (sz.height () >= 20) {
+    sz.setWidth (sz.width () + 1);
+    sz.setHeight (sz.height () + 4);
+  }
+  else if (sz.height () >= 16) {
+    sz.setHeight (sz.height () + 4);
+  }
+#else
   if (sz.height () >= 24) {
     sz.setWidth (sz.width () + 2);
     sz.setHeight (sz.height () + 6);
@@ -325,6 +462,8 @@ qt_tm_widget_rep::tweak_iconbar_size (QSize& sz) {
   else if (sz.height () >= 16) {
     sz.setHeight (sz.height () + 2);
   }
+#endif
+  //sz.setHeight ((int) floor (sz.height () * retina_scale + 0.5));
 }
 
 
@@ -332,7 +471,8 @@ qt_tm_widget_rep::tweak_iconbar_size (QSize& sz) {
  \param name A unique identifier for the window (e.g. "TeXmacs:3")
  */
 widget
-qt_tm_widget_rep::plain_window_widget (string name, command _quit) {
+qt_tm_widget_rep::plain_window_widget (string name, command _quit, int b) {
+  (void) b;
   (void) _quit; // The widget already has a command. Don't overwrite. 
   orig_name = name;
   return this;
@@ -373,7 +513,7 @@ qt_tm_widget_rep::update_visibility () {
   if ( XOR(old_statusVisibility,  new_statusVisibility) )
     mainwindow()->statusBar()->setVisible (new_statusVisibility);
 
-#ifndef Q_WS_MAC
+#ifndef Q_OS_MAC
   bool old_menuVisibility = mainwindow()->menuBar()->isVisible();
   bool new_menuVisibility = visibility[0];
 
@@ -385,8 +525,9 @@ qt_tm_widget_rep::update_visibility () {
 #ifdef UNIFIED_TOOLBAR
 
   // do modifications only if needed to reduce flicker
-  if ( XOR(old_mainVisibility,  new_mainVisibility) ||
-      XOR(old_modeVisibility,  new_modeVisibility) )
+  if (use_unified_toolbar &&
+      (XOR(old_mainVisibility,  new_mainVisibility) ||
+       XOR(old_modeVisibility,  new_modeVisibility) ))
   {
     // ensure that the topmost visible toolbar is always unified on Mac
     // (actually only for main and mode toolbars, unifying focus is not
@@ -431,12 +572,19 @@ qt_tm_widget_rep::update_visibility () {
       }
     }
   }
+  else {
+    bool old_menuVisibility = mainwindow()->menuBar()->isVisible();
+    bool new_menuVisibility = visibility[0];
+
+    if ( XOR(old_menuVisibility,  new_menuVisibility) )
+      mainwindow()->menuBar()->setVisible (new_menuVisibility);
+  }
 #endif // UNIFIED_TOOLBAR
 #undef XOR
-  {
+  if (tm_style_sheet == "" && use_mini_bars) {
     QFont f = leftLabel->font();
     int fs = as_int (get_preference ("gui:mini-fontsize", QTM_MINI_FONTSIZE));
-    f.setPixelSize (fs > 0 ? fs : QTM_MINI_FONTSIZE);
+    f.setPointSize (qt_zoom (fs > 0 ? fs : QTM_MINI_FONTSIZE));
     leftLabel->setFont(f);
     rightLabel->setFont(f);
   }
@@ -611,7 +759,7 @@ qt_tm_widget_rep::send (slot s, blackbox val) {
       ASSERT (is_nil (val), "type mismatch");
       if (!is_nil (quit))
         quit ();
-      needs_update ();
+      the_gui->need_update ();
     }
       break;
     case SLOT_FULL_SCREEN:
@@ -708,48 +856,26 @@ qt_tm_widget_rep::query (slot s, int type_id) {
 
 void
 qt_tm_widget_rep::install_main_menu () {
+  if (main_menu_widget == waiting_main_menu_widget) return;
   main_menu_widget = waiting_main_menu_widget;
-  QMenu* src = main_menu_widget->get_qmenu();
+  QList<QAction*>* src = main_menu_widget->get_qactionlist();
   if (!src) return;
-
-    // REMARK: We do not want the menubar shared across windows as suggested
-    // in http://doc.qt.nokia.com/4.7/qmainwindow.html#menuBar
-    // e.g. :
-    //
-    //     QMenuBar *dest = new QMenuBar(0);
-    //     mainwindow()->setMenuBar(dest);
-    //
-    // as the default behavior on MacOS. The main reason is that in TeXmacs
-    // different windows can have different main menus so that it is indeed
-    // appropriate to change the main menu as the window focus changes.
-    // So we kindly ask to each window to give us its own menu and we install
-    // there our actions.
-    // So we do:
-  
   QMenuBar* dest = mainwindow()->menuBar();
-  
-    // and everything is fine.
-  
-    // Also please note that we have to do the replacement and not simply
-    // install the menu returned by get_qmenu() since the main menu there
-    // could contain some default items appropriate for the given OS (like the
-    // service menu on MacOS) which are not present in our menu widget.
-    // UPDATE: But replaceActions() *does remove* all actions (!?)
-    // It looks more like the reason is that we have to "flatten out" the first
-    // level of the QMenu returned... ?
-  
-  replaceActions (dest, src);
-  QList<QAction*> list = dest->actions();
-  for (int i = 0; i < list.count(); i++) {
-    QAction* a = list[i];
+  dest->clear();
+  for (int i = 0; i < src->count(); i++) {
+    QAction* a = (*src)[i];
     if (a->menu()) {
+      //TRICK: Mac native QMenuBar accepts only menus which are already populated
+      // this will cause a problem for us, since menus are lazy and populated only after triggering
+      // this is the reason we add a dummy action before inserting the menu
+      a->menu()->addAction("native menubar trick");
+      dest->addAction(a->menu()->menuAction());
       QObject::connect (a->menu(),         SIGNAL (aboutToShow()),
                         the_gui->gui_helper, SLOT (aboutToShowMainMenu()));
       QObject::connect (a->menu(),         SIGNAL (aboutToHide()),
                         the_gui->gui_helper, SLOT (aboutToHideMainMenu()));
     }
   }
-
 }
 
 void
@@ -803,9 +929,9 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
       check_type_void (index, s);
     {
       main_icons_widget = concrete (w);
-      QMenu* m = main_icons_widget->get_qmenu();
-      if (m) {
-        replaceButtons (mainToolBar, m);
+      QList<QAction*>* list = main_icons_widget->get_qactionlist();
+      if (list) {
+        replaceButtons (mainToolBar, list);
         update_visibility();
       }
     }
@@ -815,9 +941,9 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
       check_type_void (index, s);
     {
       mode_icons_widget = concrete (w);
-      QMenu* m = mode_icons_widget->get_qmenu();
-      if (m) {
-        replaceButtons (modeToolBar, m);
+      QList<QAction*>* list = mode_icons_widget->get_qactionlist();
+      if (list) {
+        replaceButtons (modeToolBar, list);
         update_visibility();
       }
     }
@@ -826,11 +952,28 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
     case SLOT_FOCUS_ICONS:
       check_type_void (index, s);
     {
-      focus_icons_widget = concrete (w);
-      QMenu* m = focus_icons_widget->get_qmenu();
-      if (m) {
-        replaceButtons (focusToolBar, m);
-        update_visibility();
+      bool can_update = true;
+#if (QT_VERSION >= 0x050000)
+      // BUG:
+      // there is a problem with updateActions  which apparently
+      // reset a running input method in Qt5.
+      //
+      // This is (probably) also relate to
+      // bug #47338 [CJK] input disappears immediately
+      // see http://lists.gnu.org/archive/html/texmacs-dev/2017-09/msg00000.html
+      
+      // HACK: we just disable the focus bar updating while preediting.
+      // This seems enough since the other toolbars are not usually updated
+      // while performing an input method keyboard sequence
+      if (canvas()) can_update = !canvas()->isPreediting();
+#endif
+      if (can_update) {
+        focus_icons_widget = concrete (w);
+        QList<QAction*>* list = focus_icons_widget->get_qactionlist();
+        if (list) {
+          replaceButtons (focusToolBar, list);
+          update_visibility();
+        }
       }
     }
       break;
@@ -839,9 +982,9 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
       check_type_void (index, s);
     {   
       user_icons_widget = concrete (w);
-      QMenu* m = user_icons_widget->get_qmenu();
-      if (m) {
-        replaceButtons (userToolBar, m);
+      QList<QAction*>* list = user_icons_widget->get_qactionlist();
+      if (list) {
+        replaceButtons (userToolBar, list);
         update_visibility();
       }
     }
@@ -853,10 +996,7 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
       side_tools_widget = concrete (w);
       QWidget* new_qwidget = side_tools_widget->as_qwidget();
       QWidget* old_qwidget = sideTools->widget();
-      if (old_qwidget) {
-        old_qwidget->setParent (0);
-        old_qwidget->deleteLater();
-      }
+      if (old_qwidget) old_qwidget->deleteLater();
       sideTools->setWidget (new_qwidget);
       update_visibility();
       new_qwidget->show();
@@ -869,10 +1009,7 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
       bottom_tools_widget = concrete (w);
       QWidget* new_qwidget = bottom_tools_widget->as_qwidget();
       QWidget* old_qwidget = bottomTools->widget();
-      if (old_qwidget) {
-        old_qwidget->setParent (0);
-//        old_qwidget->deleteLater();
-      }
+      if (old_qwidget) old_qwidget->deleteLater();
       bottomTools->setWidget (new_qwidget);
       update_visibility();
       new_qwidget->show();
@@ -894,36 +1031,52 @@ qt_tm_widget_rep::write (slot s, blackbox index, widget w) {
   }
 }
 
+void set_standard_style_sheet (QWidget* w);
+
 void
 qt_tm_widget_rep::set_full_screen(bool flag) {
   full_screen = flag;
   QWidget *win = mainwindow()->window();  
   if (win) {
     if (flag ) {
+      QPalette pal;
+      pal.setColor(QPalette::Mid, QColor (0, 0, 0));
+      mainwindow()->setPalette(pal);
+      mainwindow()->setStyleSheet ("* { background: #000000; }");
 #ifdef UNIFIED_TOOLBAR
-      //HACK: we disable unified toolbar since otherwise
-      //  the application will crash when we return to normal mode
-      // (bug in Qt? present at least with 4.7.1)
-      mainwindow()->setUnifiedTitleAndToolBarOnMac(false);
-      mainwindow()->centralWidget()->layout()->setContentsMargins(0,0,0,0);
+      if (use_unified_toolbar) {
+        //HACK: we disable unified toolbar since otherwise
+        //  the application will crash when we return to normal mode
+        // (bug in Qt? present at least with 4.7.1)
+        mainwindow()->setUnifiedTitleAndToolBarOnMac(false);
+        mainwindow()->centralWidget()->layout()->setContentsMargins(0,0,0,0);
+      }
 #endif
 //      mainwindow()->window()->setContentsMargins(0,0,0,0);
       //win->showFullScreen();
-       win->setWindowState(win->windowState() ^ Qt::WindowFullScreen);
-    } else {
+       win->setWindowState(win->windowState() | Qt::WindowFullScreen);
+    }
+    else {
+      QPalette pal;
+      QColor bgcol= to_qcolor (tm_background);
+      pal.setColor (QPalette::Mid, bgcol);
+      mainwindow()->setPalette(pal);
+      set_standard_style_sheet (mainwindow());
       bool cache = visibility[0];
       visibility[0] = false;
       update_visibility();
 //      win->showNormal();
-      win->setWindowState(win->windowState() ^ Qt::WindowFullScreen);
+      win->setWindowState(win->windowState() & ~Qt::WindowFullScreen);
 
       visibility[0] = cache;
       update_visibility();
 #ifdef UNIFIED_TOOLBAR
-      mainwindow()->centralWidget()->layout()->setContentsMargins (0,1,0,0);
-      //HACK: we reenable unified toolbar (see above HACK) 
-      //  the application will crash when we return to normal mode
-      mainwindow()->setUnifiedTitleAndToolBarOnMac(true);
+      if (use_unified_toolbar) {
+        mainwindow()->centralWidget()->layout()->setContentsMargins (0,1,0,0);
+        //HACK: we reenable unified toolbar (see above HACK) 
+        //  the application will crash when we return to normal mode
+        mainwindow()->setUnifiedTitleAndToolBarOnMac(true);
+      }
 #endif
     }
   }
@@ -976,7 +1129,7 @@ qt_tm_embedded_widget_rep::send (slot s, blackbox val) {
       ASSERT (is_nil (val), "type mismatch");
       if (!is_nil (quit))
         quit ();
-      needs_update ();
+      the_gui->need_update ();
     }
       break;
       
