@@ -358,6 +358,12 @@ widget vertical_list (array<widget> a) {
   wid->add_children (a);
   return abstract (wid);
 }
+widget division_widget (string name, widget w) {
+  qt_widget wid = qt_ui_element_rep::create (qt_widget_rep::division_widget,
+                                             name, w);
+  wid->add_child (w);
+  return abstract (wid);
+}
 widget aligned_widget (array<widget> lhs, array<widget> rhs, SI hsep, SI vsep,
                        SI lpad, SI rpad) {
   qt_widget wid = qt_ui_element_rep::create (qt_widget_rep::aligned_widget,
@@ -547,7 +553,7 @@ widget tree_view_widget (command cmd, tree data, tree actions) {
   //// Widgets which are not strictly required by TeXmacs have void implementations
 
 widget empty_widget () { NOT_IMPLEMENTED("empty_widget"); return widget(); }
-widget extend (widget w, array<widget> a) { (void) a; return w; }
+widget extend_widget (widget w, array<widget> a) { (void) a; return w; }
 widget wait_widget (SI width, SI height, string message) {
   (void) width; (void) height; (void) message; return widget();
 }
