@@ -476,7 +476,7 @@ TeXmacs_main (int argc, char** argv) {
              (s == "-x") || (s == "-execute") ||
              (s == "-log-file") ||
              (s == "-build-manual") ||
-             (s == "-reference-suite") || (s == "-test-suite")) {}
+             (s == "-reference-suite") || (s == "-test-suite")) {i++;}
   }
   if (install_status == 1) {
     if (DEBUG_STD) debug_boot << "Loading welcome message...\n";
@@ -665,7 +665,7 @@ main (int argc, char** argv) {
   immediate_options (argc, argv);
   load_user_preferences ();
   string theme= get_user_preference ("gui theme", "default");
-#ifdef OS_MACOS
+#if defined(OS_MACOS) && !defined(__arm64__)
   if (theme == "default") theme= "";  
 #else
   if (theme == "default") theme= "light";
