@@ -16,6 +16,7 @@
 #include "merge_sort.hpp"
 #include "iterator.hpp"
 #include "boot.hpp"
+#include "drd_std.hpp"
 
 int geometry_w= 800, geometry_h= 600;
 int geometry_x= 0  , geometry_y= 0;
@@ -24,6 +25,11 @@ widget texmacs_window_widget (widget wid, tree geom);
 widget make_menu_widget (object menu);
 void refresh_size (widget wid, bool exact);
 
+static int last_window_handle= 0;
+static hashmap<int,widget> window_table (NULL);
+static hashmap<tree,path> window_by_name;
+static time_t refresh_time= 0;
+
 /******************************************************************************
 * User preference management concerning the geometry of windows
 ******************************************************************************/
@@ -70,8 +76,7 @@ static bool
 resize_accept (int old_w, int old_h, int w, int h) {
   if (old_w == w && old_h == h) return false;
 #ifdef QTTEXMACS
-  if (old_w == w && old_h == h + 16) return false;
-  if (old_w == w && old_h == h - 40) return false;
+  if (old_w == w && old_h - 80 <= h && h <= old_h + 80) return false;
 #endif
   return true;
 }
@@ -117,6 +122,7 @@ get_preferred_size (string name, SI& ww, SI& hh) {
     int h= as_int (get_user_preference ("height " * name));
     ww= w * PIXEL;
     hh= h * PIXEL;
+    //cout << "Size " << name << ": " << w << ", " << h << "\n";
   }
 }
 
@@ -133,7 +139,7 @@ tm_window_rep::tm_window_rep (widget wid2, tree geom):
   menu_current (object ()), menu_cache (widget ()),
   text_ptr (NULL)
 {
-  zoomf= get_server () -> get_default_zoom_factor ();
+  zoomf= retina_zoom * get_server () -> get_default_zoom_factor ();
 }
 
 tm_window_rep::tm_window_rep (tree doc, command quit):
@@ -143,7 +149,8 @@ tm_window_rep::tm_window_rep (tree doc, command quit):
   menu_current (object ()), menu_cache (widget ()),
   text_ptr (NULL)
 {
-  zoomf= get_server () -> get_default_zoom_factor ();
+  (void) doc;
+  zoomf= retina_zoom * get_server () -> get_default_zoom_factor ();
 }
 
 tm_window_rep::~tm_window_rep () {
@@ -167,6 +174,10 @@ texmacs_window_widget (widget wid, tree geom) {
     w= as_int (geom[0]);
     h= as_int (geom[1]);
   }
+  if (w == 800 && h == 600) {
+    w *= retina_zoom;
+    h *= retina_zoom;
+  }
   gui_root_extents (W, H); W /= PIXEL; H /= PIXEL;
   if (x < 0) x= W + x + 1 - w;
   if (y < 0) y= H + y + 1 - h;
@@ -190,8 +201,12 @@ texmacs_window_widget (widget wid, tree geom) {
 
 class close_embedded_command_rep: public command_rep {
   tm_view vw;
+  url name;
+  int win_id;
 public:
-  close_embedded_command_rep (tm_view vw2): vw (vw2) {}
+  close_embedded_command_rep (tm_view vw2, url n2, int win2):
+    vw (vw2), name (n2), win_id (win2) {
+      window_by_name(name->t)= path (win_id, window_by_name[name->t]); }
   void apply ();
   tm_ostream& print (tm_ostream& out) {
     return out << "Close_Embedded widget command"; }
@@ -200,6 +215,7 @@ public:
 void
 close_embedded_command_rep::apply () {
   //cout << "Destroy " << vw->buf->buf->name << "\n";
+  ASSERT (!is_nil(vw->ed), "embedded command acting on deleted editor");
   url foc= abstract_window (vw->ed->mvw->win);
   if (is_none (foc)) {
     array<url> a= windows_list ();
@@ -211,6 +227,7 @@ close_embedded_command_rep::apply () {
   tm_window win= vw->win;
   ASSERT (N (buffer_to_views (vw->buf->buf->name)) == 1,
           "invalid cloned embedded TeXmacs widget");
+  window_by_name(name->t)= remove<int> (window_by_name[name->t], win_id);
   remove_buffer (vw->buf->buf->name);
   //cout << "Deleted buffer\n";
   tm_delete (win);
@@ -218,8 +235,26 @@ close_embedded_command_rep::apply () {
 }
 
 command
-close_embedded_command (tm_view vw) {
-  return tm_new<close_embedded_command_rep> (vw);
+close_embedded_command (tm_view vw, url name, int win) {
+  return tm_new<close_embedded_command_rep> (vw, name, win);
+}
+
+static path
+filter_existing (path wins) {
+  if (is_nil (wins)) return wins;
+  if (window_table->contains (wins->item))
+    return path (wins->item, filter_existing (wins->next));
+  return filter_existing (wins->next);
+}
+
+path
+window_search (url name) {
+  return filter_existing (window_by_name[name->t]);
+}
+
+bool
+is_embedded_buffer (url name) {
+  return !is_nil (window_search (name));
 }
 
 /******************************************************************************
@@ -249,11 +284,16 @@ enrich_embedded_document (tree body, tree style) {
     for (int i=0; i+2<N(orig); i+=2)
       if (is_atomic (orig[i]))
         initial (orig[i]->label)= orig[i+1];
+  initial (DPI)= "720";
+  initial (ZOOM_FACTOR)= (retina_zoom==1? "1.2": "1.8");
+  initial ("no-zoom")= "true";
   tree doc (DOCUMENT);
   doc << compound ("TeXmacs", TEXMACS_VERSION);
   doc << style; //compound ("style", style);
   doc << compound ("body", body);
   doc << compound ("initial", make_collection (initial));
+  if (initial->contains ("project"))
+    doc << compound ("project", initial ["project"]);
   return doc;
 }
 
@@ -272,7 +312,8 @@ texmacs_input_widget (tree doc, tree style, url wname) {
   set_scrollable (win->wid, vw->ed);
   vw->ed->cvw= win->wid.rep;
   vw->ed->mvw= curvw;
-  return wrapped_widget (win->wid, close_embedded_command (vw));
+  command close_cmd= close_embedded_command (vw, name, last_window_handle);
+  return wrapped_widget (win->wid, close_cmd);
 }
 
 /******************************************************************************
@@ -316,9 +357,17 @@ tm_window_rep::refresh () {
 * Menus
 ******************************************************************************/
 
+bool menu_caching= true;
+
 bool
 tm_window_rep::get_menu_widget (int which, string menu, widget& w) {
+  drd_info old_drd= the_drd;
+  if (!is_none (window_to_view (id))) {
+    tm_view vw= concrete_view (window_to_view (id));
+    if (vw != NULL) the_drd= vw->ed->drd;
+  }
   object xmenu= call ("menu-expand", eval ("'" * menu));
+  the_drd= old_drd;
   //cout << "xmenu= " << xmenu << "\n";
   if (menu_cache->contains (xmenu)) {
     //if (menu_current[which] == xmenu) cout << "Same " << menu << "\n";
@@ -332,7 +381,9 @@ tm_window_rep::get_menu_widget (int which, string menu, widget& w) {
   //cout << "Compute " << menu << "\n";
   object umenu= eval ("'" * menu);
   w= make_menu_widget (umenu);
-  menu_cache (xmenu)= w;
+  if (menu_caching)
+    if (as_bool (call ("cache-menu?", xmenu)))
+      menu_cache (xmenu)= w;
   return true;
 }
 
@@ -429,13 +480,13 @@ tm_window_rep::get_bottom_tools_flag (int which) {
 
 void
 tm_window_rep::set_window_zoom_factor (double zoom) {
-  zoomf= zoom;
-  ::set_zoom_factor (wid, zoom);
+  zoomf= retina_zoom * zoom;
+  ::set_zoom_factor (wid, zoomf);
 }
 
 double
 tm_window_rep::get_window_zoom_factor () {
-  return zoomf;
+  return zoomf / retina_zoom;
 }
 
 void
@@ -531,6 +582,7 @@ tm_window_rep::interactive (string name, string type, array<string> def,
 
 void
 tm_window_rep::interactive_return () {
+  if (text_ptr == NULL) return;
   *text_ptr= get_interactive_input (wid);
   text_ptr= NULL;
   set_interactive_mode (false);
@@ -541,25 +593,13 @@ tm_window_rep::interactive_return () {
 * Other top level windows
 ******************************************************************************/
 
-static hashmap<int,widget> window_table (NULL);
-static time_t refresh_time= 0;
-
 int
 window_handle () {
   static int window_next= 1;
+  last_window_handle= window_next;
   return window_next++;
 }
 
-void
-window_create (int win, widget wid, string name, bool plain) {
-  widget pww;
-  if (plain)
-    pww= plain_window_widget (wid, name);
-  else
-    pww= popup_window_widget (wid, name);
-  window_table (win)= pww;
-}
-
 void
 window_create (int win, widget wid, string name, command quit) {
   widget pww;
@@ -567,6 +607,27 @@ window_create (int win, widget wid, string name, command quit) {
   window_table (win)= pww;
 }
 
+void
+window_create_plain (int win, widget wid, string name) {
+  widget pww;
+  pww= plain_window_widget (wid, name);
+  window_table (win)= pww;
+}
+
+void
+window_create_popup (int win, widget wid, string name) {
+  widget pww;
+  pww= popup_window_widget (wid, name);
+  window_table (win)= pww;
+}
+
+void
+window_create_tooltip (int win, widget wid, string name) {
+  widget pww;
+  pww= tooltip_window_widget (wid, name);
+  window_table (win)= pww;
+}
+
 /*
 FIXME: this old implementation does not work in the presence
 of texmacs_input widgets.  The current hack remedies this situation
