@@ -71,6 +71,12 @@ SI change_box_rep::sup_lo_base (int l) {
   return big_flag? bs[0]->sup_lo_base (l): box_rep::sup_lo_base (l); }
 SI change_box_rep::sup_hi_lim  (int l) {
   return big_flag? bs[0]->sup_hi_lim (l): box_rep::sup_hi_lim (l); }
+SI change_box_rep::wide_correction (int mode) {
+  return bs[0]->wide_correction (mode); }
+
+void change_box_rep::get_bracket_extents (SI& lo, SI& hi) {
+  if (sx(0) == 0 && sy(0) == 0) bs[0]->get_bracket_extents (lo, hi);
+  else box_rep::get_bracket_extents (lo, hi); }
 
 SI change_box_rep::get_leaf_offset (string search) {
   return sx1(0) + bs[0]->get_leaf_offset (search); }
@@ -99,27 +105,56 @@ change_box_rep::graphical_select (SI x1, SI y1, SI x2, SI y2) {
 ******************************************************************************/
 
 struct move_box_rep: public change_box_rep {
+  SI dx, dy;
   move_box_rep (path ip, box b, SI x, SI y, bool fl1, bool fl2);
   int get_type () { return MOVE_BOX; }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
   operator tree () { return tree (TUPLE, "move", (tree) bs[0]); }
 };
 
 move_box_rep::move_box_rep (path ip, box b, SI x, SI y, bool fl1, bool fl2):
-  change_box_rep (ip, fl1, fl2)
+  change_box_rep (ip, fl1, fl2), dx (x), dy (y)
 {
   insert (b, x, y);
   position ();
   finalize ();
 }
 
+box
+move_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return move_box (ip, body, dx, dy, child_flag, big_flag);
+}
+
+box
+move_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return move_box (ip, body, dx, dy, child_flag, big_flag);
+}
+
+struct move_delimiter_box_rep: public move_box_rep {
+  SI bot, top;
+  move_delimiter_box_rep (path ip, box b, SI x, SI y, SI bot2, SI top2):
+    move_box_rep (ip, b, x, y, false, true), bot (bot2), top (top2) {}
+  operator tree () {
+    return tree (TUPLE, "move_delimiter", (tree) bs[0]); }
+  void get_bracket_extents (SI& lo, SI& hi) {
+    lo= bot; hi= top; }
+};
+
 struct shift_box_rep: public change_box_rep {
+  SI dx, dy;
   shift_box_rep (path ip, box b, SI x, SI y, bool fl1, bool fl2);
   int get_type () { return MOVE_BOX; }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
   operator tree () { return tree (TUPLE, "shift", (tree) bs[0]); }
+  void get_bracket_extents (SI& lo, SI& hi);
 };
 
 shift_box_rep::shift_box_rep (path ip, box b, SI x, SI y, bool fl1, bool fl2):
-  change_box_rep (ip, fl1, fl2)
+  change_box_rep (ip, fl1, fl2), dx (x), dy (y)
 {
   insert (b, x, y);
   position ();
@@ -128,6 +163,28 @@ shift_box_rep::shift_box_rep (path ip, box b, SI x, SI y, bool fl1, bool fl2):
   finalize ();
 }
 
+box
+shift_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return move_box (ip, body, dx, dy, child_flag, big_flag);
+}
+
+box
+shift_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return move_box (ip, body, dx, dy, child_flag, big_flag);
+}
+
+void
+shift_box_rep::get_bracket_extents (SI& lo, SI& hi) {
+  if (sx(0) == 0 && sy(0) == 0) {
+    bs[0]->get_bracket_extents (lo, hi);
+    lo += dy;
+    hi += dy;
+  }
+  else box_rep::get_bracket_extents (lo, hi);
+}
+
 /******************************************************************************
 * Resizing boxes
 ******************************************************************************/
@@ -136,6 +193,8 @@ struct resize_box_rep: public change_box_rep {
   resize_box_rep (path ip, box b, SI x1, SI y1, SI x2, SI y2,
 		  bool child_flag, bool adjust);
   operator tree () { return tree (TUPLE, "resize", (tree) bs[0]); }
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
+  gr_selections graphical_select (SI x, SI y, SI dist);
 };
 
 resize_box_rep::resize_box_rep (
@@ -150,13 +209,56 @@ resize_box_rep::resize_box_rep (
   finalize ();
 }
 
-struct vcorrect_box_rep: public change_box_rep {
-  vcorrect_box_rep (path ip, box b, SI top_cor, SI bot_cor);
-  operator tree () { return tree (TUPLE, "vcorrect", (tree) bs[0]); }
+gr_selections
+resize_box_rep::graphical_select (SI x, SI y, SI dist) {
+  gr_selections res;
+  int i, n= subnr();
+  for (i=n-1; i>=0; i--)
+    res << bs[i]->graphical_select (x- sx(i), y- sy(i), dist);
+  return res;
+}
+
+struct vresize_box_rep: public change_box_rep {
+  vresize_box_rep (path ip, box b, SI y1, SI y2);
+  operator tree () { return tree (TUPLE, "vresize", (tree) bs[0]); }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
 };
 
-vcorrect_box_rep::vcorrect_box_rep (path ip, box b, SI top_cor, SI bot_cor):
-  change_box_rep (ip, false, false)
+vresize_box_rep::vresize_box_rep (path ip, box b, SI Y1, SI Y2):
+  change_box_rep (ip, false)
+{
+  insert (b, 0, 0);
+  position ();
+  x1= b->x1; y1= Y1;
+  x2= b->x2; y2= Y2;
+  finalize ();
+}
+
+box
+vresize_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return vresize_box (ip, body, y1, y2);
+}
+
+box
+vresize_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return vresize_box (ip, body, y1, y2);
+}
+
+struct vcorrect_box_rep: public change_box_rep {
+  SI top_cor, bot_cor;
+  vcorrect_box_rep (path ip, box b, SI top_cor, SI bot_cor);
+  operator tree () { return tree (TUPLE, "vcorrect", (tree) bs[0]); }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
+};
+
+vcorrect_box_rep::vcorrect_box_rep (path ip, box b, SI top_cor2, SI bot_cor2):
+  change_box_rep (ip, false, false), top_cor (top_cor2), bot_cor (bot_cor2)
 {
   insert (b, 0, -top_cor);
   position ();
@@ -165,6 +267,18 @@ vcorrect_box_rep::vcorrect_box_rep (path ip, box b, SI top_cor, SI bot_cor):
   finalize ();
 }
 
+box
+vcorrect_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return vcorrect_box (ip, body, top_cor, bot_cor);
+}
+
+box
+vcorrect_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return vcorrect_box (ip, body, top_cor, bot_cor);
+}
+
 /******************************************************************************
 * Transformed boxes
 ******************************************************************************/
@@ -189,6 +303,7 @@ public:
   void post_display (renderer &ren);
   //cursor find_cursor (path bp);
   //selection find_selection (path lbp, path rbp);
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
 };
 
 transformed_box_rep::transformed_box_rep (path ip, box b, frame fr2):
@@ -225,7 +340,10 @@ struct effect_box_rep: public change_box_rep {
 public:
   effect_box_rep (path ip, array<box> bs, tree eff);
   operator tree () { return tree (TUPLE, "effect", eff_t); }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
   void redraw (renderer ren, path p, rectangles& l);
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
 };
 
 effect_box_rep::effect_box_rep (path ip, array<box> bs, tree eff2):
@@ -252,10 +370,29 @@ effect_box_rep::effect_box_rep (path ip, array<box> bs, tree eff2):
   finalize ();
 }
 
+box
+effect_box_rep::adjust_kerning (int mode, double factor) {
+  int n= N(bs);
+  array<box> adj (n);
+  for (int i=0; i<n; i++)
+    adj[i]= bs[i]->adjust_kerning (mode, factor);
+  return effect_box (ip, adj, eff_t);
+}
+
+box
+effect_box_rep::expand_glyphs (int mode, double factor) {
+  int n= N(bs);
+  array<box> adj (n);
+  for (int i=0; i<n; i++)
+    adj[i]= bs[i]->expand_glyphs (mode, factor);
+  return effect_box (ip, adj, eff_t);
+}
+
 extern int nr_painted;
 
 void
 effect_box_rep::redraw (renderer ren, path p, rectangles& l) {
+  (void) p; (void) l;
   if (((nr_painted&15) == 15) && gui_interrupted (true)) return;
   ren->move_origin (x0, y0);
   array<picture> pics (subnr ());
@@ -295,6 +432,7 @@ public:
   void pre_display (renderer &ren);
   void post_display (renderer &ren);
   selection find_selection (path lbp, path rbp);
+  void get_bracket_extents (SI& lo, SI& hi) { lo= y1; hi= y2; }
 };
 
 clip_box_rep::clip_box_rep (
@@ -334,12 +472,18 @@ clip_box_rep::find_selection (path lbp, path rbp) {
 ******************************************************************************/
 
 struct repeat_box_rep: public change_box_rep {
-  repeat_box_rep (path ip, box b, box repeat, SI xoff);
+  box repeat;
+  SI xoff;
+  bool under;
+  repeat_box_rep (path ip, box b, box repeat, SI xoff, bool under);
   operator tree () { return tree (TUPLE, "repeat", (tree) bs[0]); }
+  int reindex (int i, int item, int n);
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
 };
 
-repeat_box_rep::repeat_box_rep (path ip, box b, box repeat, SI xoff):
-  change_box_rep (ip, false)
+repeat_box_rep::repeat_box_rep (path ip, box b, box r2, SI xoff2, bool u2):
+  change_box_rep (ip, false), repeat (r2), xoff (xoff2), under (u2)
 {
   insert (b, 0, 0);
   position ();
@@ -354,6 +498,10 @@ repeat_box_rep::repeat_box_rep (path ip, box b, box repeat, SI xoff):
     for (i=i1; i<i2; i++) {
       box bb= move_box (decorate_right (ip), repeat, 0, 0);
       insert (bb, i*width-xoff, 0);
+      x3= min (x3, i*width-xoff + repeat->x3);
+      x4= max (x4, i*width-xoff + repeat->x4);
+      y3= min (y3, repeat->y3);
+      y4= max (y4, repeat->y4);
     }
   }
 
@@ -362,24 +510,48 @@ repeat_box_rep::repeat_box_rep (path ip, box b, box repeat, SI xoff):
   x2= b->x2; y2= b->y2;
 }
 
+int
+repeat_box_rep::reindex (int i, int item, int n) {
+  if (under) return (i<n? i+1: 0);
+  else return change_box_rep::reindex (i, item, n);
+}
+
+box
+repeat_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return repeat_box (ip, body, repeat, xoff, under);
+}
+
+box
+repeat_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return repeat_box (ip, body, repeat, xoff, under);
+}
+
 /******************************************************************************
 * cell boxes for tables
 ******************************************************************************/
 
 struct cell_box_rep: public change_box_rep {
+  SI    X0, Y0;
   SI    bl, br, bb, bt;
   brush fg, bg, old_bg;
   cell_box_rep (path ip, box b, SI x0, SI y0, SI x1, SI y1, SI x2, SI y2,
 		SI bl, SI br, SI bb, SI bt, brush fg, brush bg);
   operator tree () { return tree (TUPLE, "cell", (tree) bs[0]); }
+  box  adjust_kerning (int mode, double factor);
+  box  expand_glyphs (int mode, double factor);
+  void get_cell_extents (SI& l, SI& r);
+  box  adjust_cell_geometry (SI dx, SI dl, SI dr);
   void pre_display (renderer &ren);
   void post_display (renderer &ren);
 };
 
 cell_box_rep::cell_box_rep (
-  path ip, box b, SI X0, SI Y0, SI X1, SI Y1, SI X2, SI Y2,
+  path ip, box b, SI X0b, SI Y0b, SI X1, SI Y1, SI X2, SI Y2,
   SI Bl, SI Br, SI Bb, SI Bt, brush Fg, brush Bg):
   change_box_rep (ip, false),
+  X0 (X0b), Y0 (Y0b),
   bl (Bl<<1), br (Br<<1), bb (Bb<<1), bt (Bt<<1),
   fg (Fg), bg (Bg)
 {
@@ -397,6 +569,36 @@ cell_box_rep::cell_box_rep (
   finalize ();
 }
 
+box
+cell_box_rep::adjust_kerning (int mode, double factor) {
+  box nb= bs[0]->adjust_kerning (mode & (~TABLE_CELL), factor);
+  SI  d = nb->w() - bs[0]->w();
+  if ((mode & TABLE_CELL) != 0) d= 0;
+  return cell_box (ip, nb, X0, Y0, x1, y1, x2 + d, y2,
+                   bl>>1, br>>1, bb>>1, bt>>1, fg, bg);
+}
+
+box
+cell_box_rep::expand_glyphs (int mode, double factor) {
+  box nb= bs[0]->expand_glyphs (mode & (~TABLE_CELL), factor);
+  SI  d = nb->w() - bs[0]->w();
+  if ((mode & TABLE_CELL) != 0) d= 0;
+  return cell_box (ip, nb, X0, Y0, x1, y1, x2 + d, y2,
+                   bl>>1, br>>1, bb>>1, bt>>1, fg, bg);
+}
+
+void
+cell_box_rep::get_cell_extents (SI& l, SI& r) {
+  l= sx1 (0);
+  r= sx2 (0);
+}
+
+box
+cell_box_rep::adjust_cell_geometry (SI dx, SI dl, SI dr) {
+  return cell_box (ip, bs[0], X0+dx, Y0, x1+dl, y1, x2+dr, y2,
+                   bl>>1, br>>1, bb>>1, bt>>1, fg, bg);
+}
+
 void
 cell_box_rep::pre_display (renderer& ren) {
   SI l= bl, r= br, b= bb, t= bt;
@@ -473,40 +675,6 @@ public:
   }
 };
 
-/******************************************************************************
-* action boxes
-******************************************************************************/
-
-struct action_box_rep: public change_box_rep {
-  tree    filter; // which events are accepted ?
-  command cmd;    // command to be executed
-  path    vip;    // store this location before execution
-  action_box_rep (path ip, box b, tree f, command c, bool ch, path vip);
-  operator tree () { return tree (TUPLE, "action", bs[0]); }
-  tree action (tree t, SI x, SI y, SI delta);
-};
-
-action_box_rep::action_box_rep (
-  path ip, box b, tree filter2, command cmd2, bool child_flag, path vip2):
-  change_box_rep (ip, child_flag), filter (filter2), cmd (cmd2), vip (vip2)
-{
-  insert (b, 0, 0);
-  position ();
-  left_justify ();
-  finalize ();
-}
-
-tree
-action_box_rep::action (tree t, SI x, SI y, SI delta) {
-  if (t == filter) {
-    call ("action-set-path", reverse (vip));
-    // FIXME: we should also reset the action path
-    cmd ();
-    return "done";
-  }
-  return change_box_rep::action (t, x, y, delta);
-}
-
 /******************************************************************************
 * locus boxes
 ******************************************************************************/
@@ -519,9 +687,10 @@ struct locus_box_rep: public change_box_rep {
   locus_box_rep (path ip, box b, list<string> ids, SI pixel);
   locus_box_rep (path ip, box b, list<string> ids, SI pixel, string _rep, string _anchor);
   operator tree () { return tree (TUPLE, "locus"); }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
   void loci (SI x, SI y, SI delta, list<string>& ids2, rectangles& rs);
   void post_display (renderer &ren);
-
 };
 
 locus_box_rep::locus_box_rep (path ip, box b, list<string> ids2, SI pixel2):
@@ -546,18 +715,29 @@ locus_box_rep::locus_box_rep (path ip, box b, list<string> ids2, SI pixel2, stri
   finalize ();
 }
 
+box
+locus_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return locus_box (ip, body, ids, pixel, ref, anchor);
+}
+
+box
+locus_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return locus_box (ip, body, ids, pixel, ref, anchor);
+}
 
 void
 locus_box_rep::loci (SI x, SI y, SI delta, list<string>& l, rectangles& rs) {
   bs[0]->loci (x, y, delta, l, rs);
   l = l * ids;
-  rs= rs * outline (rectangles (rectangle (x1, y1, x2, y2)), pixel);
+  rs= rs * outlines (rectangles (rectangle (x1, y1, x2, y2)), pixel);
 }
 
 void
 locus_box_rep::post_display (renderer &ren) {
-  if (ref!="") ren->href(ref, x1, y1, x2, y2);
-  if (anchor!="") ren->anchor(anchor, x1, y1);
+  if (ref != "") ren->href (ref, x1, y1, x2, y2);
+  if (anchor != "") ren->anchor (anchor, x1, y1, x2, y2);
 }
 
 /******************************************************************************
@@ -566,15 +746,18 @@ locus_box_rep::post_display (renderer &ren) {
 
 struct tag_box_rep: public change_box_rep {
   tree keys;
-  tag_box_rep (path ip, box b, tree keys);
+  path tip;
+  tag_box_rep (path ip, path tip, box b, tree keys);
   operator tree () { return tree (TUPLE, "tag", bs[0]); }
+  box adjust_kerning (int mode, double factor);
+  box expand_glyphs (int mode, double factor);
   tree tag (tree t, SI x, SI y, SI delta);
   void collect_page_numbers (hashmap<string,tree>& h, tree page);
   path find_tag (string name);
 };
 
-tag_box_rep::tag_box_rep (path ip, box b, tree keys2):
-  change_box_rep (ip, false), keys (keys2)
+tag_box_rep::tag_box_rep (path ip, path tip2, box b, tree keys2):
+  change_box_rep (ip, false), keys (keys2), tip (tip2)
 {
   insert (b, 0, 0);
   position ();
@@ -582,6 +765,18 @@ tag_box_rep::tag_box_rep (path ip, box b, tree keys2):
   finalize ();
 }
 
+box
+tag_box_rep::adjust_kerning (int mode, double factor) {
+  box body= bs[0]->adjust_kerning (mode, factor);
+  return tag_box (ip, tip, body, keys);
+}
+
+box
+tag_box_rep::expand_glyphs (int mode, double factor) {
+  box body= bs[0]->expand_glyphs (mode, factor);
+  return tag_box (ip, tip, body, keys);
+}
+
 void
 tag_box_rep::collect_page_numbers (hashmap<string,tree>& h, tree page) {
   for (int i=0; i<N(keys); i++)
@@ -593,10 +788,32 @@ path
 tag_box_rep::find_tag (string search) {
   for (int i=0; i<N(keys); i++)
     if (keys[i]->label == search)
-      return reverse (descend_decode (ip, 1));
+      return reverse (descend_decode (tip, 1));
   return path ();
 }
 
+/******************************************************************************
+* note boxes
+******************************************************************************/
+
+struct note_box_rep: public change_box_rep {
+  note_box_rep (path ip, box b, box note, SI nx, SI ny);
+  operator tree () { return tree (TUPLE, "note", bs[0], bs[1]); }
+};
+
+note_box_rep::note_box_rep (path ip, box b, box note, SI nx, SI ny):
+  change_box_rep (ip, false)
+{
+  insert (b, 0, 0);
+  insert (note, nx, ny);
+  position ();
+  finalize ();
+  x1= b->x1;
+  y1= b->y1;
+  x2= b->x2;
+  y2= b->y2;
+}
+
 /******************************************************************************
 * text_at boxes
 ******************************************************************************/
@@ -698,6 +915,11 @@ move_box (path ip, box b, SI x, SI y, bool child_flag, bool big_flag) {
   return tm_new<move_box_rep> (ip, b, x, y, child_flag, big_flag);
 }
 
+box
+move_delimiter_box (path ip, box b, SI x, SI y, SI bot, SI top) {
+  return tm_new<move_delimiter_box_rep> (ip, b, x, y, bot, top);
+}
+
 box
 shift_box (path ip, box b, SI x, SI y, bool child_flag, bool big_flag) {
   return tm_new<shift_box_rep> (ip, b, x, y, child_flag, big_flag);
@@ -709,6 +931,11 @@ resize_box (path ip, box b, SI x1, SI y1, SI x2, SI y2,
   return tm_new<resize_box_rep> (ip, b, x1, y1, x2, y2, child_flag, adjust);
 }
 
+box
+vresize_box (path ip, box b, SI y1, SI y2) {
+  return tm_new<vresize_box_rep> (ip, b, y1, y2);
+}
+
 box
 transformed_box (path ip, box b, frame fr) {
   return tm_new<transformed_box_rep> (ip, b, fr);
@@ -736,8 +963,8 @@ vcorrect_box (path ip, box b, SI top_cor, SI bot_cor) {
 }
 
 box
-repeat_box (path ip, box ref, box repeat, SI xoff) {
-  return tm_new<repeat_box_rep> (ip, ref, repeat, xoff);
+repeat_box (path ip, box ref, box repeat, SI xoff, bool under) {
+  return tm_new<repeat_box_rep> (ip, ref, repeat, xoff, under);
 }
 
 box
@@ -754,16 +981,6 @@ remember_box (path ip, box b) {
   return tm_new<remember_box_rep> (ip, b);
 }
 
-box
-action_box (path ip, box b, tree filter, command cmd, bool ch, path vip) {
-  return tm_new<action_box_rep> (ip, b, filter, cmd, ch, vip);
-}
-
-box
-action_box (path ip, box b, tree filter, command cmd, bool ch) {
-  return tm_new<action_box_rep> (ip, b, filter, cmd, ch, decorate ());
-}
-
 box
 locus_box (path ip, box b, list<string> ids, SI pixel) {
   return tm_new<locus_box_rep> (ip, b, ids, pixel);
@@ -775,8 +992,13 @@ locus_box (path ip, box b, list<string> ids, SI pixel, string ref, string anchor
 }
 
 box
-tag_box (path ip, box b, tree keys) {
-  return tm_new<tag_box_rep> (ip, b, keys);
+tag_box (path ip, path tip, box b, tree keys) {
+  return tm_new<tag_box_rep> (ip, tip, b, keys);
+}
+
+box
+note_box (path ip, box b, box note, SI nx, SI ny) {
+  return tm_new<note_box_rep> (ip, b, note, nx, ny);
 }
 
 box
