@@ -40,7 +40,7 @@ void
 concater_rep::marker (path ip) {
   if (is_decoration (ip)) ip= path (0, ip);
   string fn_name= "cmr";
-  int sz= script (env->fn_size, env->index_level);
+  int sz= env->get_script_size (env->fn_size, env->index_level);
   font gfn (tex_font (fn_name, sz, (int) (env->magn*env->dpi)));
   box b= text_box (ip->next, ip->item, "", gfn, blue);
   a << line_item (MARKER_ITEM, OP_SKIP, b, HYPH_INVALID);
@@ -48,7 +48,7 @@ concater_rep::marker (path ip) {
 
 void
 concater_rep::ghost (string s, path ip) {
-  ghost (s, ip, blue);
+  ghost (s, ip, env->src_tag_col);
 }
 
 void
@@ -69,7 +69,7 @@ concater_rep::ghost (string s, path ip, color col) {
     else if (s[0] == '{') { fn_name= "cmsy"; s= "f"; }
     else if (s[0] == '}') { fn_name= "cmsy"; s= "g"; }
   }
-  int sz= script (env->fn_size, env->index_level);
+  int sz= env->get_script_size (env->fn_size, env->index_level);
   font gfn (tex_font (fn_name, sz, (int) (env->magn*env->dpi)));
   box b= text_box (decorate (ip), 0, s, gfn, col);
   array<box> bs (1);
@@ -84,23 +84,30 @@ concater_rep::flag_ok (string s, path ip, color col) {
   SI h= 4*env->fn->wfn/5;
   int r, g, b, a;
   get_rgb_color (col, r, g, b, a);
-  r= 255- (255 - r)/6;
-  g= 255- (255 - g)/6;
-  b= 255- (255 - b)/6;
+  //r= 255- (255 - r)/6;
+  //g= 255- (255 - g)/6;
+  //b= 255- (255 - b)/6;
+  a= a/6;
   color light= rgb_color (r, g, b, a);
   int info= env->info_level;
-  if (info == INFO_MINIMAL || info == INFO_SHORT) {
+  if (info == INFO_MINIMAL || info == INFO_SHORT || info == INFO_SHORT_PAPER) {
     box infob= info_box (dip, h, pencil (col, env->fn->wline), light);
-    box specb= specific_box (ip, infob, false, env->fn);
-    print (specb);
+    if (info == INFO_SHORT_PAPER) {
+      box b= resize_box (ip, infob, 0, 0, 0, env->fn->yx);
+      print (b);
+    }
+    else {
+      box specb= specific_box (ip, infob, "screen", env->fn);
+      print (specb);
+    }
   }
   else if (info == INFO_DETAILED || info == INFO_PAPER) {
-    int sz= script (env->fn_size, env->index_level+2);
+    int sz= env->get_script_size (env->fn_size, env->index_level+2);
     font gfn (tex_font ("ecrm", sz, (int) (env->magn*env->dpi)));
     box textb= text_box (decorate (ip), 0, s, gfn, col);
     box flagb= flag_box (dip, textb, h, pencil (col, env->fn->wline), light);
     if (info == INFO_DETAILED) {
-      box specb= specific_box (ip, flagb, false, env->fn);
+      box specb= specific_box (ip, flagb, "screen", env->fn);
       print (specb);
     }
     else {
@@ -161,6 +168,7 @@ concater_rep::typeset (tree t, path ip) {
 
   if (!is_accessible (ip)) {
     path ip2= obtain_ip (t);
+    //if (ip2 != ip) cout << t << ", " << ip << " -> " << ip2 << "\n";
     if (ip2 != path (DETACHED))
       ip= ip2;
   }
@@ -199,6 +207,9 @@ concater_rep::typeset (tree t, path ip) {
   case RIGID:
     typeset_rigid (t, ip);
     break;
+  case HGROUP:
+    typeset_hgroup (t, ip);
+    break;
   case HIDDEN:
     //(void) env->exec (t);
     if (N(t) != 1) typeset_error (t, ip);
@@ -250,7 +261,10 @@ concater_rep::typeset (tree t, path ip) {
     typeset_clipped (t, ip);
     break;
   case REPEAT:
-    typeset_repeat (t, ip);
+    typeset_repeat (t, ip, false);
+    break;
+  case VAR_REPEAT:
+    typeset_repeat (t, ip, true);
     break;
   case _FLOAT:
     typeset_float (t, ip);
@@ -267,6 +281,15 @@ concater_rep::typeset (tree t, path ip) {
   case DBOX:
     typeset_decorated_box (t, ip);
     break;
+  case LINE_NOTE:
+    typeset_line_note (t, ip);
+    break;
+  case PAGE_NOTE:
+    typeset_page_note (t, ip);
+    break;
+  case IF_PAGE_BREAK:
+    typeset_if_page_break (t, ip);
+    break;
 
   case WITH_LIMITS:
     with_limits (LIMITS_DISPLAY);
@@ -315,6 +338,10 @@ concater_rep::typeset (tree t, path ip) {
   case PAGE_BREAK:
   case VAR_NO_PAGE_BREAK:
   case NO_PAGE_BREAK:
+  case VAR_NO_BREAK_HERE:
+  case NO_BREAK_HERE:
+  case NO_BREAK_START:
+  case NO_BREAK_END:
   case VAR_NEW_PAGE:
   case NEW_PAGE:
   case VAR_NEW_DPAGE:
@@ -335,7 +362,7 @@ concater_rep::typeset (tree t, path ip) {
     typeset_large (t, ip, LEFT_BRACKET_ITEM, OP_OPENING_BRACKET, "<left-");
     break;
   case MID:
-    typeset_large (t, ip, MIDDLE_BRACKET_ITEM, OP_MIDDLE_BRACKET, "<mid-");
+    typeset_wide_middle (t, ip);
     break;
   case RIGHT:
     typeset_large (t, ip, RIGHT_BRACKET_ITEM, OP_CLOSING_BRACKET, "<right-");
@@ -408,17 +435,23 @@ concater_rep::typeset (tree t, path ip) {
   case ASSIGN:
     typeset_assign (t, ip);
     break;
+  case PROVIDE:
+    typeset_provide (t, ip);
+    break;
   case WITH:
     typeset_with (t, ip);
     break;
   case PROVIDES:
     typeset_executable (t, ip);
     break;
+  case VALUE:
+    typeset_value (t, ip);
+    break;
   case QUOTE_VALUE:
     typeset_inactive (t, ip);
     break;
-  case VALUE:
-    typeset_value (t, ip);
+  case OR_VALUE:
+    typeset_or_value (t, ip);
     break;
   case MACRO:
     typeset_inactive (t, ip);
@@ -439,8 +472,6 @@ concater_rep::typeset (tree t, path ip) {
     typeset_inactive (t, ip);
     break;
   case GET_LABEL:
-    typeset_executable (t, ip);
-    break;
   case GET_ARITY:
     typeset_executable (t, ip);
     break;
@@ -450,6 +481,12 @@ concater_rep::typeset (tree t, path ip) {
   case EVAL_ARGS:
     typeset_eval_args (t, ip);
     break;
+  case NEW_THEME:
+  case COPY_THEME:
+  case APPLY_THEME:
+  case SELECT_THEME:
+    typeset_executable (t, ip);
+    break;
   case MARK:
     typeset_mark (t, ip);
     break;
@@ -487,9 +524,15 @@ concater_rep::typeset (tree t, path ip) {
   case EXTERN:
     typeset_rewrite (t, ip);
     break;
-  case INCLUDE:
+  case VAR_INCLUDE:
     typeset_include (t, ip);
     break;
+  case INCLUDE:
+    typeset_compound (t, ip);
+    break;
+  case WITH_PACKAGE:
+    typeset_rewrite (t, ip);
+    break;
   case USE_PACKAGE:
   case USE_MODULE:
     typeset_executable (t, ip);
@@ -526,14 +569,19 @@ concater_rep::typeset (tree t, path ip) {
   case TRANSLATE:
   case CHANGE_CASE:
   case FIND_FILE:
+  case FIND_FILE_UPWARDS:
   case IS_TUPLE:
   case LOOK_UP:
+  case OCCURS_INSIDE:
   case EQUAL:
   case UNEQUAL:
   case LESS:
   case LESSEQ:
   case GREATER:
   case GREATEREQ:
+  case BLEND:
+  case RGB_COLOR:
+  case RGB_ACCESS:
   case BOX_INFO:
   case FRAME_DIRECT:
   case FRAME_INVERSE:
@@ -561,15 +609,30 @@ concater_rep::typeset (tree t, path ip) {
   case SPC_LENGTH:
   case XSPC_LENGTH:
   case PAR_LENGTH:
+  case PAW_LENGTH:
   case PAG_LENGTH:
   case GW_LENGTH:
   case GH_LENGTH:
   case GU_LENGTH:
   case TMPT_LENGTH:
   case PX_LENGTH:
+  case GUIPX_LENGTH:
+  case LCORNER_LENGTH:
+  case BCORNER_LENGTH:
+  case RCORNER_LENGTH:
+  case TCORNER_LENGTH:
+  case MS_LENGTH:
+  case S_LENGTH:
+  case MSEC_LENGTH:
+  case SEC_LENGTH:
+  case MIN_LENGTH:
+  case HR_LENGTH:
     typeset_executable (t, ip);
     break;
 
+  case FILTER_STYLE:
+    typeset_executable (t, ip);
+    break;
   case STYLE_ONLY:
   case VAR_STYLE_ONLY:
   case ACTIVE:
@@ -605,6 +668,8 @@ concater_rep::typeset (tree t, path ip) {
   case LINK:
   case URL:
   case SCRIPT:
+  case OBSERVER:
+  case FIND_ACCESSIBLE:
     typeset_inactive (t, ip);
     break;
   case HLINK:
@@ -615,16 +680,25 @@ concater_rep::typeset (tree t, path ip) {
     typeset_set_binding (t, ip);
     break;
   case GET_BINDING:
+  case HAS_BINDING:
     typeset_executable (t, ip);
     break;
+  case HIDDEN_BINDING:
+    break;
   case LABEL:
   case REFERENCE:
   case PAGEREF:
     typeset_compound (t, ip);
     break;
+  case GET_ATTACHMENT:
+    typeset_executable (t, ip);
+    break;
   case WRITE:
     typeset_write (t, ip);
     break;
+  case TOC_NOTIFY:
+    typeset_toc_notify (t, ip);
+    break;
 
   case TUPLE:
   case ATTR:
@@ -646,7 +720,17 @@ concater_rep::typeset (tree t, path ip) {
   case FLAG:
     typeset_flag (t, ip);
     break;
+  case HYPHENATE_AS:
+    typeset_hyphenate_as (t, ip);
+    break;
 
+  case ANIM_STATIC:
+  case ANIM_DYNAMIC:
+  case MORPH:
+  case ANIM_TIME:
+  case ANIM_PORTION:
+    typeset_executable (t, ip);
+    break;
   case ANIM_COMPOSE:
     typeset_anim_compose (t, ip);
     break;
@@ -656,6 +740,9 @@ concater_rep::typeset (tree t, path ip) {
   case ANIM_CONSTANT:
     typeset_anim_constant (t, ip);
     break;
+  case ANIM_ACCELERATE:
+    typeset_anim_accelerate (t, ip);
+    break;
   case ANIM_TRANSLATE:
     typeset_anim_translate (t, ip);
     break;
@@ -690,6 +777,9 @@ concater_rep::typeset (tree t, path ip) {
   case MATH_AT:
     typeset_math_at (t, ip);
     break;
+  case DOCUMENT_AT:
+    typeset_document_at (t, ip);
+    break;
   case _POINT:
     typeset_point (t, ip);
     break;
@@ -714,6 +804,12 @@ concater_rep::typeset (tree t, path ip) {
   case CSPLINE:
     typeset_cspline (t, ip);
     break;
+  case BEZIER:
+  case CBEZIER:
+  case SMOOTH:
+  case CSMOOTH:
+    typeset_bezier (t, ip);
+    break;
   case FILL:
     typeset_fill (t, ip);
     break;
@@ -734,6 +830,11 @@ concater_rep::typeset (tree t, path ip) {
   case EFF_MOVE:
   case EFF_MAGNIFY:
   case EFF_BUBBLE:
+  case EFF_CROP:
+  case EFF_TURBULENCE:
+  case EFF_FRACTAL_NOISE:
+  case EFF_HATCH:
+  case EFF_DOTS:
   case EFF_GAUSSIAN:
   case EFF_OVAL:
   case EFF_RECTANGULAR:
@@ -742,6 +843,9 @@ concater_rep::typeset (tree t, path ip) {
   case EFF_OUTLINE:
   case EFF_THICKEN:
   case EFF_ERODE:
+  case EFF_DEGRADE:
+  case EFF_DISTORT:
+  case EFF_GNAW:
   case EFF_SUPERPOSE:
   case EFF_ADD:
   case EFF_SUB:
@@ -752,8 +856,11 @@ concater_rep::typeset (tree t, path ip) {
   case EFF_NORMALIZE:
   case EFF_MONOCHROME:
   case EFF_COLOR_MATRIX:
+  case EFF_GRADIENT:
   case EFF_MAKE_TRANSPARENT:
   case EFF_MAKE_OPAQUE:
+  case EFF_RECOLOR:
+  case EFF_SKIN:
     typeset_inactive (t, ip);
     break;
 
@@ -763,6 +870,9 @@ concater_rep::typeset (tree t, path ip) {
   case ORNAMENT:
     typeset_ornament (t, ip);
     break;
+  case ART_BOX:
+    typeset_art_box (t, ip);
+    break;
 
   default:
     if (L(t) < START_EXTENSIONS) print (test_box (ip));
@@ -803,6 +913,7 @@ typeset_as_concat (edit_env env, tree t, path ip) {
   ccc->typeset (t, ip);
   ccc->finish ();
   array<line_item> a= ccc->a;
+  tm_delete (ccc);
 
   int i, n=N(a);
   if (n == 0) return empty_box (ip); // FIXME: n=0 should never happen
@@ -818,7 +929,31 @@ typeset_as_concat (edit_env env, tree t, path ip) {
   }
   box b= concat_box (ip, items, spc);
 
+  return b;
+}
+
+box
+typeset_as_concat (edit_env env, tree t, path ip, array<line_item>& a) {
+  concater ccc= tm_new<concater_rep> (env, true);
+  ccc->typeset (t, ip);
+  ccc->finish ();
+  a= ccc->a;
   tm_delete (ccc);
+
+  int i, n=N(a);
+  if (n == 0) return empty_box (ip); // FIXME: n=0 should never happen
+  array<box> items (n);
+  array<SI>  spc (n);
+  if (n>0) {
+    spc[0]=0;
+    for (i=0; i<n-1; i++) {
+      items[i]  = a[i]->b;
+      spc  [i+1]= a[i]->spc->def;
+    }
+    items[i]= a[i]->b;
+  }
+  box b= concat_box (ip, items, spc);
+
   return b;
 }
 
